package jumbo.xml;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.FileDialog;
import java.awt.Font;
import java.awt.Label;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintWriter;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.sun.java.swing.ButtonGroup;
import com.sun.java.swing.Icon;
import com.sun.java.swing.ImageIcon;
import com.sun.java.swing.JComponent;
import com.sun.java.swing.JCheckBoxMenuItem;
import com.sun.java.swing.JFrame;
import com.sun.java.swing.JLabel;
import com.sun.java.swing.JMenu;
import com.sun.java.swing.JMenuBar;
import com.sun.java.swing.JMenuItem;
import com.sun.java.swing.JPanel;
import com.sun.java.swing.JRadioButtonMenuItem;
import com.sun.java.swing.JScrollPane;
import com.sun.java.swing.JSplitPane;
import com.sun.java.swing.JTree;
import com.sun.java.swing.JTextField;
import com.sun.java.swing.ToolTipManager;
import com.sun.java.swing.UIManager;

import com.sun.java.swing.tree.DefaultTreeModel;
import com.sun.java.swing.tree.TreePath;

import jumbo.xml.gui.XSplitPanePanel;
import jumbo.xml.gui.TabbedPane;

/** an XML-based Tree view, (a forerunner of the DOM). Unlike XNode
(which represents the model), XTree is a JTree which is a view of the
tree. The tree model can be extracted from XTree if required. (If this
is all new to you, don't worry - it took me some time - ignore it for
now.)
<BR>
XTree can be constructed from the rootNode of a tree based on XNodes.
(The example main() program shows the logic:
<PRE>
		XNode rootNode = new XNode("root");
		XNode xNode = new XNode("Chapter a");
		XNode x0 = null;
		rootNode.add(xNode);
		xNode.addAttribute("ID", "chapA", Attribute.CDATA);
		xNode.addAttribute("foo", "bar", Attribute.CDATA);
		xNode.add(x0 = new XNode("Section a1"));
		x0.add(new PCDATANode("This is a section a1 string"));
		xNode.add(x0 = new XNode("Section a2"));
		x0.add(new PCDATANode("This is a section a2 string"));
		xNode.add(x0 = new XNode("Section a3"));
		x0.add(new PCDATANode("This is a section a3 string"));
		xNode = new XNode("Chapter b");
		rootNode.add(xNode);
		xNode.add(new XNode("Section b1"));
		xNode.add(new XNode("Section b2"));
		xNode.add(new XNode("Section b3"));

		XTree xTree = new XTree(rootNode);
		xTree.display();
</PRE>
The tree is fully contained in the XNode structure and the XTree is
primarily used to render it. Run:
<PRE>
java  jumbo.xml.XTree foo 
</PRE>
to show the effect.
<BR>
The default display uses an XSplitPanePanel with the XTree on the
left, but the XTree can be embdedded in any JComponent if required.
The tree has several options set by XTreeOptions, which in principle 
allows for TreeCellRenderer. Unfortunately although I got this working
nicely, it's rebelled for a bit and so shouldn't be used.
<BR>
Program and logic is highly derivative from SwingSet example.
 
@author P.Murray-Rust, 1998
*/
public class XTree extends JTree implements ActionListener {

/** the tree model (default if not specified) */
	protected DefaultTreeModel treeModel = null;
/** the root of the tree */
	protected XNode rootNode = null;						   
/** a container for the XNodePanel */
	protected JComponent xNodePanelContainer = null;
/** the panel for rendering of the XNode components (e.g. Attributes) */
	protected XNodePanel xNodePanel = null;				   
/** the options for tree display */
	protected XTreeOptions options;
/** the frame in which all is displayed */
	protected JFrame jFrame = null;
/** the message area in it */
	protected JLabel jMessage = new JLabel("Message");
/** a box for the URL */
	protected JPanel jURLPanel = new JPanel();
/** textField for the URL */
	protected JTextField jURLTextField = null;
/** the URL */
	protected String urlString = null;
/** the current parserClassName */
	protected String parserClassName = SAX.parserTypeStrings[SAX.defaultParserType];
/** the current output format; default is TAGGED event stream */
	protected int outputFormat = XNode.TAGGED;
/** the current help tree */
	protected XTree helpTree = null;
/** the helpfile - don't know how to locate this... */
	protected static String helpUrl = "jumbohelp.xml";

/** all nodes analysis */
	public final static int ALLNODES	= 0;
/** element analysis */
	public final static int ELEMENT		= 1;
/** PCDATA analysis */
	public final static int PCDATA		= 2;
//	public final static int nAnalysisTypes = PCDATA+1;
//	public final static String[] analysisTypeStrings = {
//		"All Nodes",
//		"Elements",
//		"PCDATA (Strings)",
//	};

//	JCheckBoxMenuItem contentDisplayCheckBox[] = new JCheckBoxMenuItem[XNode.nComponentTypes];
	JRadioButtonMenuItem contentDisplayRadioButton[] = new JRadioButtonMenuItem[XNode.nComponentTypes];
	JRadioButtonMenuItem parserRadioButton[] = new JRadioButtonMenuItem[SAX.nParserTypes];

// holds cut-and-paste as a node+subtree
	XNode clipboardNode = null;

// keep a record of the root nodes. This is so we can find the tree
// that any XNode belongs to.
    static Hashtable rootNodeTable = new Hashtable();

/** the simplest way of creating an XTree. 
Create the tree with a default rootNode
*/
	public XTree() {
		this(new XNode("_ROOT"));
	}

/** the next simplest way of creating an XTree. 
Create the tree using XNode.add(XNode) first.
@param XNode rootNode the root node of the tree
*/
	public XTree(XNode rootNode) {
		super(rootNode);
		treeModel = new DefaultTreeModel(rootNode);
		initialise(rootNode);
	}

/** 
Create a single node tree .
@param String name the name of the root node of the tree
@param boolean askAllowsChildren see DefaultMutableTreeNode for this. I use true so far.
*/
	public XTree(String name, boolean askAllowsChildren) {
		this(new XNode(name), askAllowsChildren);
	}


/** 
Create a single node tree .
@param String name the name of the root node of the tree
*/
	public XTree(String name) {
		this(new XNode(name));
	}

/** 
Create a tree from a root node .
@param String name the name of the root node of the tree
@param boolean askAllowsChildren see DefaultMutableTreeNode for this. I use true so far.
*/
	public XTree(XNode rootNode, boolean askAllowsChildren) {
		super(rootNode, askAllowsChildren);
		treeModel = new DefaultTreeModel(rootNode, askAllowsChildren);
		initialise(rootNode);
	}
	
	void initialise(XNode rootNode) {
		this.rootNode = rootNode;
// register the node
		rootNodeTable.put(rootNode, this);

		jFrame = new JFrame("XTree");
		jFrame.getContentPane().setLayout(new BorderLayout());
		addMenu();
// Box for URLs
		jFrame.getContentPane().add(jURLPanel, BorderLayout.NORTH);
		jURLPanel.setLayout(new BorderLayout());
		jURLPanel.add(new JLabel("Current URL: "), BorderLayout.WEST);
		jURLPanel.add(new JLabel("(edit+RET to change)"), BorderLayout.EAST);
		jURLTextField = new JTextField(80);
		jURLTextField.setText("");

		jURLPanel.add(jURLTextField, BorderLayout.CENTER);
// load URL on return
		URLActionListener al = new URLActionListener(this);
		jURLTextField.addActionListener(al);

		jFrame.getContentPane().add(jMessage, BorderLayout.SOUTH);
		options = new XTreeOptions();
		try {
			UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
	    // If you want the System L&F instead, comment out the above line and
	    // uncomment the following:
	    // UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception exc) {
			System.err.println("Error loading L&F: " + exc);
		}
		if (options.renderCell) {
			this.setCellRenderer(new XTreeCellRenderer());
		}

	}

/** find tree that an XNode belongs to (shopuld be previously registered)
*/
	public static XTree getTree(XNode node) {
		XNode rootNode = (XNode) node.getRoot();
		return (XTree) rootNodeTable.get(rootNode);
	}
	
	public void addMenu() {

        JMenuBar jmb = new JMenuBar();
        
		JMenu fileMenu = new JMenu ("File");
        JMenuItem item;
        fileMenu.add (item = new JMenuItem ("New"));
        item.addActionListener (this);
  		item.setEnabled(false);
		fileMenu.add (item = new JMenuItem ("Open File"));
        item.addActionListener (this);
//		fileMenu.add (item = new JMenuItem ("Open URL"));
//        item.addActionListener (this);
        fileMenu.addSeparator();
		fileMenu.add (item = new JMenuItem ("Save"));
        item.addActionListener (this);
		fileMenu.add (item = new JMenuItem ("Save As"));
        item.addActionListener (this);
		fileMenu.add (item = new JMenuItem ("Save Selected As"));
        item.addActionListener (this);
        fileMenu.addSeparator();
// output format
		JMenu outputMenu = new JMenu ("Output Format");
		fileMenu.add(outputMenu);
        ButtonGroup outputGroup = new ButtonGroup();
		JRadioButtonMenuItem outputRadioButton[] = 
			new JRadioButtonMenuItem[XNode.displayTypeStrings.length];
        for (int i = XNode.XMLCONTENT; i < XNode.displayTypeStrings.length; i++) {
			outputRadioButton[i] = 
				new JRadioButtonMenuItem("O:"+XNode.displayTypeStrings[i]);
			outputRadioButton[i].addActionListener (this);
			outputMenu.add(outputRadioButton[i]);
			outputGroup.add(outputRadioButton[i]);
		}
		outputRadioButton[outputFormat].setSelected(true);
        fileMenu.addSeparator();

        fileMenu.add (item = new JMenuItem ("Exit"));
        item.addActionListener (this);
        jmb.add (fileMenu);

        JMenu editMenu = new JMenu ("Edit");
        editMenu.add (item = new JMenuItem ("Cut"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Copy"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Paste Youngest Child"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Paste Elder Sibling"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Paste Younger Sibling"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Show Clipboard"));
        item.addActionListener (this);
        editMenu.addSeparator();
        editMenu.add (item = new JMenuItem ("Add Dummy Child"));
        item.addActionListener (this);
        editMenu.addSeparator();
        editMenu.add (item = new JMenuItem ("Add Numbering"));
        item.addActionListener (this);
        editMenu.addSeparator();
        editMenu.add (item = new JMenuItem ("Delete Whitespace"));
        item.addActionListener (this);
        editMenu.add (item = new JMenuItem ("Delete Whitespace Selected"));
        item.addActionListener (this);
        jmb.add (editMenu);

        JMenu viewMenu = new JMenu ("View");
        JMenu contentDisplayMenu = new JMenu ("Content Display");
		viewMenu.add(contentDisplayMenu);
        for (int i = XNode.XMLCONTENT; i < XNode.nComponentTypes; i++) {
//			contentDisplayCheckBox[i] = new JCheckBoxMenuItem(XNode.displayTypeStrings[i]);
//			contentDisplayCheckBox[i].addActionListener (this);
//			viewMenu.add (contentDisplayCheckBox[i]);
			contentDisplayRadioButton[i] = new JRadioButtonMenuItem(XNode.displayTypeStrings[i]);
			contentDisplayRadioButton[i].addActionListener (this);
			contentDisplayMenu.add (contentDisplayRadioButton[i]);
		}
//		contentDisplayCheckBox[XNode.CONTENT].setSelected(true);
		contentDisplayRadioButton[XNode.CONTENT].setSelected(true);

        viewMenu.addSeparator();
		viewMenu.add (item = new JMenuItem ("Structural Analysis"));
		item.addActionListener (this);

		jmb.add (viewMenu);

// Parse menu
        JMenu parseMenu = new JMenu ("Parsing");
        JMenu parsersMenu = new JMenu ("Parsers");
		parseMenu.add(parsersMenu);

        ButtonGroup parserGroup = new ButtonGroup();
        for (int i = 0; i < SAX.nParserTypes; i++) {
			parserRadioButton[i] = new JRadioButtonMenuItem(SAX.parserTypeStrings[i]);
			parserRadioButton[i].addActionListener (this);
			parsersMenu.add (parserRadioButton[i]);
			parserGroup.add(parserRadioButton[i]);
		}
		parserRadioButton[SAX.defaultParserType].setSelected(true);

		jmb.add (parseMenu);

// Help
        JMenu jumboMenu = new JMenu ("Help");

// Jumbo icon
        jumboMenu.add (item = new JMenuItem ("Jumbo"));
        item.addActionListener (this);
  		item.setEnabled(true);
		try {
			ImageIcon icon = new ImageIcon("images/jumbo.gif");
			if (icon != null) {
				item.setIcon(icon);
			}
		} catch (Exception e) {
			System.out.println("No icon: "+e);
		}

        jumboMenu.add (item = new JMenuItem ("Help"));
        item.addActionListener (this);
  		item.setEnabled(true);

		jmb.add (jumboMenu);

        jFrame.setJMenuBar (jmb);
	}

/** set the tree options by cloning an existing XTreeOptions
@param XTreeOptions tree display options
*/
	public void setOptions(XTreeOptions options) {
		this.options = new XTreeOptions(options);
	}
	
/** displays a clock cursor if true, else default.  Could be extended
to lock the frame when I know how
*/    

    public void displayWait(boolean onOff) {
        try {
            if (onOff) {
				System.out.println("<wait>");
                //this.setEnabled(false);
                this.setCursor(new Cursor(Cursor.WAIT_CURSOR));
				//this.update();
            } else {
				System.out.println("</wait>");
                //this.setEnabled(true);
                this.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
				//this.update();
            }
        } catch (Exception e) {
            System.out.println("DisplayWait: "+e);
        }
    }

/** display the tree. Can be controlled by XTreeOptions, the default being
to use an XSplitPanePanel in a JFrame. It adds a MouseListener which
can be customised for different types of clicks. It also sets toolTips.

*/
	public void display() {
		final JTree tree = this;
		if (treeModel != null) tree.setModel(treeModel);
		MouseListener ml = new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				int selRow = tree.getRowForLocation(e.getX(), e.getY());
				TreePath selPath = tree.getPathForLocation(e.getX(), e.getY());
				if(selRow != -1) {
					try {
						onClick(selRow, selPath, e);
					} catch (NullPointerException npe) {
						System.out.println("Node clicked NPE:"+npe);
						npe.printStackTrace(System.out);
					} catch (Exception ex) {
						System.out.println("Node clicked: "+ex);
						ex.printStackTrace();
					}
				}
			}								 
		};
		tree.addMouseListener(ml);

/* Enable tool tips for the tree, without this tool tips will not
  be picked up. */
		ToolTipManager.sharedInstance().registerComponent(tree);

/* Make the tree use an instance of XTreeCellRenderer for
  drawing. */
// there is some horrid bug in the display here I can't find :-(
// the icons are not rendered
		if (options.renderCell) {
			tree.setCellRenderer(new XTreeCellRenderer());
		}
		JScrollPane treePane = new JScrollPane(tree);

		xNodePanel = new XNodePanel(this);
		if (options.displayType == XTreeOptions.SPLITPANE) {
			xNodePanelContainer = new JPanel();
			xNodePanelContainer.setLayout(new BorderLayout());
			xNodePanelContainer.add(xNodePanel, BorderLayout.CENTER);
			xNodePanelContainer.add(new JLabel("XNode Panel"), BorderLayout.SOUTH);
			XSplitPanePanel xsp = new XSplitPanePanel(JSplitPane.HORIZONTAL_SPLIT, treePane, xNodePanelContainer);
			jFrame.getContentPane().add(xsp, BorderLayout.CENTER);
		} else if (options.displayType == XTreeOptions.SINGLEPANE) {
			jFrame.getContentPane().add(treePane, BorderLayout.CENTER);
		}
		jFrame.pack();
		jFrame.show();
		jFrame.setSize(600,450);
	}

/** gets root node (probably better Swing ways :-)
@return XNode the root node
*/
	public XNode getRootNode() {
		return rootNode;
	}

/** gets the (first) selected node
@return XNode the selected node
*/ 
	public XNode getSelectedNode() {
		TreePath treePath = this.getSelectionPath();
		if (treePath == null) return null;
		Object[] objs = treePath.getPath();
		if (objs == null) return null;
		Object userObject = objs[objs.length-1];
		if (userObject == null) return null;
		if (!(userObject instanceof XNode)) return null;
		return (XNode) userObject;
	}


/** gets all selected nodes (null if none)
@return Vector the selected nodes (null if none);
*/ 
	public Vector getSelectedNodes() {
		TreePath[] treePaths = this.getSelectionPaths();
		if (treePaths == null) return null;
		Vector nodeVector = new Vector();
		for (int i = 0; i < treePaths.length; i++) {
			Object[] objs = treePaths[i].getPath();
			if (objs == null) continue;
			Object userObject = objs[objs.length-1];
			if (userObject == null) continue;
			if (!(userObject instanceof XNode)) continue;
			nodeVector.addElement(userObject);
		}
		if (nodeVector.size() == 0) return null;
		return nodeVector;
	}

// process a mouse click on a node
	void onClick(int selRow, TreePath selPath, MouseEvent event) {
		if (Util.isRightMouseClick(event)) {
		} else if (event.getClickCount() == 1) {
			return;
		}
		Object[] objs = selPath.getPath();
		Object userObject = objs[objs.length-1];
		if (!(userObject instanceof XNode)) return;
		XNode xNode = (XNode) userObject;
		if (xNode == null) {
			System.out.println("Null node clicked: "+selRow);
			return;
		}
// in case this is a specialised object, else no-op
		xNode.onClick();

// single panel tree needs new window
		if (xNodePanelContainer == null) {
			JFrame displayFrame = new JFrame("XNode Panel");
			xNodePanelContainer = new JPanel();
			displayFrame.getContentPane().add(xNodePanelContainer, BorderLayout.CENTER);
			xNodePanelContainer.setLayout(new BorderLayout());
			xNodePanelContainer.add(xNodePanel, BorderLayout.CENTER);
			displayFrame.pack();
			displayFrame.show();
			displayFrame.setSize(500,500);
		}
		xNodePanel.addXNode(xNode);
	}

/** gets the panel into which XNodes can be displayed
*/
	public XNodePanel getXNodePanel() {
		return xNodePanel;
	}

/** doesn't yet work
@param XNode node the node to highlight
*/
	public void selectAndHighlight(XNode node) {
		Object[] nodePath = node.getPath();
		for (int i = 0; i < nodePath.length; i++) {
			System.out.println(""+nodePath[i].getClass()+nodePath[i].toString());
		}
		TreePath treePath = new TreePath(nodePath);
		this.setSelectionPath(treePath);
		this.scrollPathToVisible(treePath);
	}

// process a click on a node
    public void actionPerformed (ActionEvent e) {
        try {
			String s = e.getActionCommand();
// File menu
			if (s.equals("Open File")) {
				loadFile();
/**--
			} else if (s.equals("Open URL")) {
				loadURL();
--*/
			} else if (s.equals("Save")) {	 
				saveXML(rootNode, outputFormat, urlString);
			} else if (s.equals("Save As")) {	 
				saveXML(rootNode, outputFormat, null);
			} else if (s.equals("Save Selected As")) {
				saveXML(this.getSelectedNodes(), outputFormat, null);
			} else if (s.equals("Exit")) {
				System.exit(1);

// Edit menu
			} else if (s.equals("Copy")) {
				XNode xNode = this.getSelectedNode();
				if (xNode != null) {
					clipboardNode = new XNode(xNode);
				}
			} else if (s.equals("Cut")) {
				clipboardNode = this.getSelectedNode();
				if (clipboardNode != null) {
					XNode parent = (XNode) clipboardNode.getParent();
					clipboardNode.removeFromParent();
					resetTree(parent);
				}
			} else if (s.equals("Paste Youngest Child")) {
				if (clipboardNode != null) {
					XNode xNode = this.getSelectedNode();
					if (xNode != null) {
						XNode newNode = new XNode(clipboardNode);
						xNode.add(newNode);
						resetTree(newNode);
					}
				}
			} else if (s.equals("Paste Younger Sibling")) {
				if (clipboardNode != null) {
					XNode xNode = this.getSelectedNode();
					if (xNode != null) {
						XNode parentNode = (XNode) xNode.getParent();
						if (parentNode != null) {
							int index = parentNode.getIndex(xNode);
							XNode newNode = new XNode(clipboardNode);
							parentNode.insert(newNode, index+1);
							resetTree(newNode);
						}
					}
				}
			} else if (s.equals("Paste Elder Sibling")) {
				if (clipboardNode != null) {
					XNode xNode = this.getSelectedNode();
					if (xNode != null) {
						XNode parentNode = (XNode) xNode.getParent();
						if (parentNode != null) {
							int index = parentNode.getIndex(xNode);
							XNode newNode = new XNode(clipboardNode);
							parentNode.insert(newNode, index);
							resetTree(newNode);
						}
					}
				}
			} else if (s.equals("Add Dummy Child")) {
				Vector nodeVector = this.getSelectedNodes();
				if (nodeVector != null) {
					for (int i = 0; i < nodeVector.size(); i++) {
						XNode node = (XNode) nodeVector.elementAt(i);
						XNode newNode = new XNode("#DUMMY", null);
						node.add(newNode);
						resetTree(newNode);
					}
				}
			} else if (s.equals("Add Numbering")) {
				numberNodes(rootNode, "");
				resetTree();

			} else if (s.equals("Show Clipboard")) {
				if (clipboardNode != null) {
					new XTree(clipboardNode).display();
				}
			} else if (s.equals("Delete Whitespace")) {
				deleteWhiteSpace();

//View Menu
			} else if (s.equals("Structural Analysis")) {
				doAnalysis();

// Help menu
			} else if (s.equals("Help")) {
				if (helpTree == null) {
					helpUrl = Util.makeAbsoluteURL(helpUrl);
					System.out.println("Help from: "+helpUrl);
					SAX sax = new SAX(helpUrl, SAX.defaultParserClassName);
					helpTree = sax.getTree();
					helpTree.display();
				}

			} else {
// output format
				for (int i = XNode.XMLCONTENT; i < XNode.displayTypeStrings.length; i++) {
					if (s.equals("O:"+XNode.displayTypeStrings[i])) {
						outputFormat = i;
						return;
					}
				}

// no action required for display Types...
				for (int i = XNode.XMLCONTENT; i < XNode.nComponentTypes; i++) {
					if (s.equals(XNode.displayTypeStrings[i])) {
						return;
					}
				}

//Parse Menu
// set current parser if checked
				for (int i = 0; i < SAX.nParserTypes; i++) {
					if (s.equals(SAX.parserTypeStrings[i])) {
						parserClassName = s;
						return;
					}
				}

			}
		} catch (Exception ex) {
			System.out.println("Action Exception: "+ex);
		}
    }
	
/** numbers nodes by levels (e.g. 1.2.3). omitting PCDATA and PI. The numbering
adds digits to the title field. At present it cannot be saved (although we might
generate a TITLE attribute??).  
@param XNode node the node to start numbering at (only the node and the subtree)
@param String prefix. Numbers will be of form Prefix.1.2.3: title, where title was the original title
*/
	public void numberNodes(XNode node, String prefix) {
		int level = 0;
		Enumeration childs = node.children();
		int count = 1;
		if (childs != null) {
			while (childs.hasMoreElements()) {
				XNode child = (XNode) childs.nextElement();
				if (child instanceof PCDATANode || child instanceof PINode) {
				} else {
					if (child.title == null) child.title = "";
					String prefix1 = prefix+"."+(count++);
					child.title = prefix1+": "+child.title;
					numberNodes((XNode) child, prefix1);
				}
			}
		}
	}

/** saves node to a file selected by the user.
@param XNode node the node and its subtree are written to file
@param int control XNode.TAGGED, XNode.PRETTY, etc.
@exception Exception almost anything - probably mainly java.io
*/
	public void saveXML(XNode node, int control, String urlString) 
		throws Exception {
		if (node == null) return;
		if (urlString == null) {
			urlString = getURLString();
		}
		if (urlString == null) return;
		System.out.println("Writing file: "+urlString);
		this.displayWait(true);
        PrintWriter pw = new PrintWriter(new FileOutputStream(urlString));
		node.writeEventStream(pw, control, 0);
		pw.close();
		System.out.println("Wrote file: "+urlString);
		this.displayWait(false);
	}

/** saves nodes to a file selected by the user.	Output format ignored at present
@param Vector nodeVector vector of nodes to be saved;
@exception Exception almost anything - probably mainly java.io
*/
	public void saveXML(Vector nodeVector, int control, String urlString) 
		throws Exception {
		if (nodeVector == null) return;
		if (urlString == null) {
			urlString = getURLString();
		}
		if (urlString == null) return;
		System.out.println("Writing file: "+urlString);
		this.displayWait(true);
        PrintWriter pw = new PrintWriter(new FileOutputStream(urlString));
		pw.println("<SELECTEDNODES>");
		for (int i = 0; i < nodeVector.size(); i++) {
			XNode node = (XNode) nodeVector.elementAt(i);
			pw.println(node.getEventStream(control, 0));
		}
		pw.println("</SELECTEDNODES>");
		pw.close();
		System.out.println("Wrote file: "+urlString);
		this.displayWait(false);
	}

	void loadFile() {
		urlString = getURLString();
		loadUrlString(urlString);
		jURLTextField.setText(urlString);
	}

	FileDialog fileDialog = null;
	String getURLString() {
		if (fileDialog == null) {
			fileDialog = new FileDialog(jFrame);
		}
		fileDialog.setMode(FileDialog.LOAD);
	    fileDialog.show();

	    String file = fileDialog.getFile();
	    if (file == null) {
			return null;
	    }
	    String directory = fileDialog.getDirectory();
	    File f = new File(directory, file);
		return f.toString();
	}
	
	void loadURL() {
		System.out.println("Load URL NYI");
	}

	void loadUrlString(String urlString) {
// W95
		if (!urlString.startsWith("/")) {
			urlString = "/"+urlString;
		}
		urlString = ("file:"+urlString).replace('\\', '/');
		URL url = null;
		try {
			url = new URL(urlString);
		} catch (MalformedURLException e) {return;}
		System.out.println("URL: "+url);
		this.displayWait(true);
		try {
			SAX sax = new SAX(url.toString(), parserClassName);
			XTree newTree = sax.getTree();
			rootNode = newTree.rootNode;
			resetTree();
		} catch (XException e) {
			System.out.println("Cannot make tree: "+e);
		}
		this.displayWait(false);
	}

// if node structure has been changed, reload root Node
	public void resetTree() {resetTree(null);}

/** if node structure has been changed, reload root Node 
	and highlight node
*/
	public void resetTree(XNode node) {
		treeModel = new DefaultTreeModel(rootNode);
		this.setModel(treeModel);
		xNodePanelContainer.removeAll();
		xNodePanel = new XNodePanel(this);
		xNodePanelContainer.add(xNodePanel, BorderLayout.CENTER);
		xNodePanelContainer.validate();
		if (node != null) {
			selectNode(node);
		}
	}

	void deleteWhiteSpace() {
		rootNode.deleteWhitespaceDescendants();
		resetTree();
	}

// hashtable and tree for holding element names
	Hashtable elementTable = null;
	XTree elementTree = null;
// analysis of nodes
	void doAnalysis() {
		Enumeration enum = rootNode.depthFirstEnumeration();
		Vector analysisVector = new Vector();
		rootNode.depthFirstEnumeration();
		while (enum.hasMoreElements()) {
			XNode xNode = (XNode) enum.nextElement();
			if (!(xNode instanceof PCDATANode) &&
				    !(xNode instanceof PINode)) {
				analysisVector.addElement(xNode);
			}
		}
		int nNodes = analysisVector.size();
		System.out.println("Analysis: "+nNodes);
// for elements build a tree
		XNode rootNode = new XNode("Elements");
		elementTree = new XTree(rootNode);
		elementTable = new Hashtable();
		for (int i = 0; i < nNodes; i++) {
			addToElementTree((XNode) analysisVector.elementAt(i));
		}
		elementTree.display();
	}

// for each new elementName create a new node and hashtable.
// the actual nodes go in the hashtable. The hashtable is the 
// use object for the node.
	void addToElementTree(XNode node) {
		String elementName = node.elementName;
		ElementNode elementTreeNode = null;
		XNode attributeParentNode = null;
		XNode contentParentNode = null;
		AttributeNode attributeNode = null;
// new elementName
		if (elementTable.get(elementName) == null) {
			Hashtable nodeTable = new Hashtable();
			Hashtable attributeTable = new Hashtable();
			nodeTable.put("#ATTRIBUTE", attributeTable);
			Hashtable contentTable = new Hashtable();
			nodeTable.put("#CONTENT", contentTable);
			elementTable.put(elementName, nodeTable);
			elementTreeNode = new ElementNode(elementName, nodeTable, this);
			elementTree.rootNode.add(elementTreeNode);
			nodeTable.put("#XNODE", elementTreeNode);
		}
		Hashtable nodeTable = (Hashtable) elementTable.get(elementName);
		Hashtable attributeTable = (Hashtable) nodeTable.get("#ATTRIBUTE");
		Hashtable contentTable = (Hashtable) nodeTable.get("#CONTENT");
		elementTreeNode = (ElementNode) nodeTable.get("#XNODE");

		nodeTable.put(node, "");
// and attributes
		if (node.attributeVector != null) {
			if (nodeTable.get("#ATTRIBUTEPARENTNODE") == null) {
				attributeParentNode = new XNode("Attributes", null);
				nodeTable.put("#ATTRIBUTEPARENTNODE", attributeParentNode);
				elementTreeNode.add(attributeParentNode);
			}
			attributeParentNode = (XNode) nodeTable.get("#ATTRIBUTEPARENTNODE");
			for (int i = 0; i < node.attributeVector.size(); i++) {
				Attribute attribute = (Attribute) node.attributeVector.elementAt(i);
// attributeNameTable is separate table for each attribute name in each
// element.  Holds all the nodes with given elementName/attributeName
				Hashtable attributeNameTable = null;
				if (attributeTable.get(attribute.name) == null) {
					attributeNameTable = new Hashtable();
					attributeTable.put(attribute.name, attributeNameTable);
					attributeNode = new AttributeNode(attribute.name, elementName, this);
					attributeParentNode.add(attributeNode);
// make another level of table for attribute values
					Hashtable attributeValueTable = new Hashtable();
					attributeNameTable.put("#ATTRIBUTEVALUE", attributeValueTable);
					attributeNameTable.put("#ATTRIBUTENODE", attributeNode);
				}
				attributeNameTable = (Hashtable) attributeTable.get(attribute.name);
				attributeNode = (AttributeNode) attributeNameTable.get("#ATTRIBUTENODE");
// store all nodes with this attribute name
				attributeNameTable.put(node, "");
// store nodes by attribute value
				if (attribute.value != null) {
					String valueNodeTitle = attribute.value.substring(0, Math.min(20, attribute.value.length()));
					Hashtable attributeValueTable = (Hashtable) attributeNameTable.get("#ATTRIBUTEVALUE");
					if (attributeValueTable.get(valueNodeTitle) == null) {
						attributeValueTable.put(valueNodeTitle, new Vector());
						AttributeValueNode attributeValueNode = new AttributeValueNode(valueNodeTitle, elementName, attribute.name, this);
						attributeNode.add(attributeValueNode);
					}
				    Vector valueVector = (Vector) attributeValueTable.get(valueNodeTitle);
					valueVector.addElement(node);
				}
			}
		}
// and children
		if (nodeTable.get("#CONTENTNODE") == null) {
			contentParentNode = new XNode("Content", null);
			nodeTable.put("#CONTENTNODE", contentParentNode);
			elementTreeNode.add(contentParentNode);
		}
		contentParentNode = (XNode) nodeTable.get("#CONTENTNODE");
		Enumeration childs = node.children();
		if (childs != null) {
			while (childs.hasMoreElements()) {
				XNode child = (XNode) childs.nextElement();
				String contentName = child.elementName;
// contentTable is separate table for each content in each
// element.  Holds all the nodes with given elementName/childElementName
				Hashtable contentNameTable = null;
				if (contentTable.get(contentName) == null) {
					contentNameTable = new Hashtable();
					contentTable.put(contentName, contentNameTable);
					ContentNode contentNode = new ContentNode(contentName, elementName, this);
					contentParentNode.add(contentNode);
				}
				contentNameTable = (Hashtable) contentTable.get(contentName);
				contentNameTable.put(node, "");
			}
		}
	}

/** selects a node and highlights it
@param XNode xNode the node to highlight
*/
	public void selectNode(XNode xNode) {
		if (xNode == null) return;
		TreePath treePath = new TreePath(treeModel.getPathToRoot(xNode));
		this.setSelectionPath(treePath);
		this.scrollPathToVisible(treePath);
	}

/** selects nodes and highlights them
@param Vector xNode the node to highlight
*/
	public void selectNodes(Vector nodeVector) {
		if (nodeVector == null) return;
		TreePath[] treePaths = new TreePath[nodeVector.size()];
		for (int i = 0; i < treePaths.length; i++) {
			treePaths[i] = new TreePath(treeModel.getPathToRoot((XNode)nodeVector.elementAt(i)));
			this.scrollPathToVisible(treePaths[i]);
		}
		this.setSelectionPaths(treePaths);
	}

/** search for a given element, attribute, value
@param XNode topNode Node to start at (all descendants are searched)
@param String elementName (if null matches any)
@param String attributeName (if null matches any, regardless of attributeValue)
@param String attributeValue(if null matches any)
@param int sensitivity XNode.CASE or XNode.IGNORECASE (applies to all params)
*/
	public Vector search(XNode topNode, String elementName, 
		String attributeName, String attributeValue, int sensitivity) {
		Vector nodeVector = null;
		Enumeration nodes = topNode.depthFirstEnumeration();
		if (nodes != null) nodeVector = new Vector();
		while (nodes.hasMoreElements()) {
			XNode node = (XNode) nodes.nextElement();
			if (elementName == null || 
				node.elementName.equals(elementName) ||
				(sensitivity == XNode.CASE && node.elementName.equalsIgnoreCase(elementName))) {
				nodeVector.addElement(node);
			}
		}
		int nElements = nodeVector.size();
		if (nElements == 0) return null;
		if (attributeName == null) return nodeVector;

		Vector nodeVector1 = new Vector();
		for (int i = 0; i < nElements; i++) {
			XNode node = (XNode) nodeVector.elementAt(i);
			String attValue = node.getAttributeValue(attributeName, sensitivity);
			if (attValue != null) {
				if (attributeValue == null || 
					attributeValue.equals(attValue) ||
					(sensitivity == XNode.CASE && attributeValue.equalsIgnoreCase(attValue))) {
					nodeVector1.addElement(node);
				}
			}
		}
		if (nodeVector1.size() == 0) return null;
		return nodeVector1;
	}

/** post a message to the JMessage message area
@param String message the message
*/
	public void postMessage(String message) {
		jMessage.setText(message);
	}

/** a test program, see above
*/
	public static void main(String[] args) {
		if (args.length == 0) {
			System.out.println("Usage: jumbo.xml.XTree");
			System.exit(0);
		}
		XNode rootNode = new XNode("root");
		XNode xNode = new XNode("Chapter a");
		XNode x0 = null;
		rootNode.add(xNode);
		xNode.addAttribute("ID", "chapA", Attribute.CDATA);
		xNode.addAttribute("foo", "bar", Attribute.CDATA);
		xNode.add(x0 = new XNode("Section a1"));
		x0.add(new PCDATANode("This is a section a1 string"));
		xNode.add(x0 = new XNode("Section a2"));
		x0.add(new PCDATANode("This is a section a2 string"));
		xNode.add(x0 = new XNode("Section a3"));
		x0.add(new PCDATANode("This is a section a3 string"));
		xNode.add(x0 = new XNode("Section a4"));
		x0.add(new PCDATANode("This is a section a4 string"));
		xNode.add(x0 = new XNode("Section a5"));
		x0.add(new PCDATANode("This is a section a5 string"));
		xNode.add(x0 = new XNode("Section a6"));
		x0.add(new PCDATANode("This is a section a6 string"));
		try {
			XNode.makeSubClassedNode(xNode, "jumbo.xml.TestSubClassedNode");
		} catch (XException e) {Util.bug(e);}
		xNode = new XNode("Chapter b");
		xNode.addAttribute("ID", "chapB", Attribute.CDATA);
		rootNode.add(xNode);
		xNode.add(x0 = new XNode("Section b1"));
		x0.add(new PCDATANode("This is a section b1 string"));
		xNode.add(x0 = new XNode("Section b2"));
		x0.add(new PCDATANode("This is a section b2 string"));
		try {
			xNode = XNode.makeSubClassedNode(xNode, "jumbo.xml.TestSubClassedNode");
		} catch (XException e) {Util.bug(e);}
		xNode.add(x0 = new XNode("Section b3"));
		x0.add(new PCDATANode("This is a section b3 string"));
		xNode.add(x0 = new XNode("Section b4"));
		x0.add(new PCDATANode("This is a section b4 string"));
		xNode.add(x0 = new XNode("Section b5"));
		x0.add(new PCDATANode("This is a section b5 string"));
		xNode.add(x0 = new XNode("Section b6"));
		x0.add(new PCDATANode("This is a section b6 string"));

		XTree xTree = new XTree(rootNode);
// should set cell rendering (e.g. icon) but fails
		XTreeOptions options = new XTreeOptions();
		//options.renderCell = true;
		xTree.setOptions(options);
		xTree.display();
		Vector nodeVector = xTree.search(xTree.rootNode, null, "ID", null, XNode.IGNORECASE);
		xTree.selectNodes(nodeVector);
	}
}
class ElementNode extends XNode {
	XTree xTree;
	public ElementNode() {
		super();
	}
	public ElementNode(String elementName, Object object, XTree xTree) {
		super(elementName, object);
		this.xTree = xTree;
	}
	public void onClick() {
		Hashtable nodeTable = (Hashtable) xTree.elementTable.get(this.elementName);
		Enumeration enum = nodeTable.keys();
		Vector nodeVector = new Vector();
		while (enum.hasMoreElements()) {
			Object obj = enum.nextElement();
			if (obj instanceof XNode) {
				nodeVector.addElement(obj);
			}
		}
		xTree.selectNodes(nodeVector);
	}
}
class AttributeNode extends XNode {
	XTree xTree;
	String attributeName;
	String elementTreeName;
	public AttributeNode() {
		super();
	}
	public AttributeNode(String attributeName, String elementTreeName, XTree xTree) {
		super(attributeName, elementTreeName);
		this.xTree = xTree;
		this.attributeName = attributeName;
		this.elementTreeName = elementTreeName;
	}
// not yet working
	public void onClick() {
		Hashtable nodeTable = (Hashtable) xTree.elementTable.get(this.elementTreeName);
		Hashtable attributeTable = (Hashtable) nodeTable.get("#ATTRIBUTE");
		Hashtable attributeNameTable = (Hashtable) attributeTable.get(attributeName);
		Enumeration enum = attributeNameTable.keys();
		Vector nodeVector = new Vector();
		while (enum.hasMoreElements()) {
			Object obj = enum.nextElement();
			if (obj instanceof XNode) {
				nodeVector.addElement(obj);
			}
		}
		xTree.selectNodes(nodeVector);
	}
}
class ContentNode extends XNode {
	XTree xTree;
	String contentName;
	String elementTreeName;
	public ContentNode() {
		super();
	}
	public ContentNode(String contentName, String elementTreeName, XTree xTree) {
		super(contentName, elementTreeName);
		this.xTree = xTree;
		this.contentName = contentName;
		this.elementTreeName = elementTreeName;
	}
// not yet working
	public void onClick() {
		Hashtable nodeTable = (Hashtable) xTree.elementTable.get(this.elementTreeName);
		Hashtable contentTable = (Hashtable) nodeTable.get("#CONTENT");
		Hashtable contentNameTable = (Hashtable) contentTable.get(contentName);
		Enumeration enum = contentNameTable.keys();
		Vector nodeVector = new Vector();
		while (enum.hasMoreElements()) {
			Object obj = enum.nextElement();
			if (obj instanceof XNode) {
				nodeVector.addElement(obj);
			}
		}
		xTree.selectNodes(nodeVector);
	}
}
class AttributeValueNode extends XNode {
	XTree xTree;
	String attributeName;
	String valueName;
	String elementTreeName;
	public AttributeValueNode() {
		super();
	}
	public AttributeValueNode(String valueName, String elementTreeName, String attributeName, XTree xTree) {
		super(valueName, elementTreeName);
		this.xTree = xTree;
		this.attributeName = attributeName;
		this.valueName = valueName;
		this.elementTreeName = elementTreeName;
	}
// not yet working
	public void onClick() {
		Hashtable nodeTable = (Hashtable) xTree.elementTable.get(this.elementTreeName);
		if (nodeTable == null) return;
		Hashtable attributeTable = (Hashtable) nodeTable.get("#ATTRIBUTE");
		if (attributeTable == null) return;
		Hashtable attributeNameTable = (Hashtable) attributeTable.get(attributeName);
		if (attributeNameTable == null) return;
		Hashtable attributeValueTable = (Hashtable) attributeNameTable.get("#ATTRIBUTEVALUE");
		if (attributeValueTable == null) return;
		Vector valueVector = (Vector) attributeValueTable.get(valueName);
		xTree.selectNodes(valueVector);
	}
}
class URLActionListener implements ActionListener {
	XTree xTree	= null;
	public URLActionListener(XTree xTree) {
		this.xTree = xTree;
	}
	public void actionPerformed(ActionEvent e) {
		System.out.println("E"+e);
		String urlString = xTree.jURLTextField.getText();
		System.out.println("S"+urlString);
		xTree.loadUrlString(urlString);
	}
};
